from .reachability_algorithm import ReachabilityAlgorithm
import logging
import numpy as np

logger = logging.getLogger(__name__)


class TOPPRA(ReachabilityAlgorithm):
    """Time-Optimal Path Parameterization based on Reachability
    Analysis (TOPPRA).

    Examples
    -----------
    >>> instance = algo.TOPPRA([pc_vel, pc_acc], path)
    >>> jnt_traj = instance.compute_trajectory()  # rest-to-rest motion
    >>> instance.problem_data # intermediate result

    Parameters
    ----------
    constraint_list: List[:class:`~toppra.constraint.Constraint`]
        List of constraints to which the robotic system is subjected to.
    path: :class:`.AbstractGeometricPath`
        Input geometric path.
    gridpoints: Optional[np.ndarray]
        Gridpoints for discretization of the geometric path. The start
        and end points must agree with the geometric path's
        `path_interval`. If omited a gridpoint will be automatically
        selected.
    solver_wrapper: str, optional
        Name of the solver wrapper to use. Possible value are:

        - 'seidel'
        - 'hotqpoases'

        For more details see the solverwrappers documentation.

    parametrizer: str, optional
        Name of the output parametrizer to use.

    Notes
    -----
    In addition to the given constraints, there are additional
    constraints on the solutions enforced by the solver-wrapper.
    Therefore, different parametrizations are returned for different
    solver wrappers. However, the difference should be very small,
    especially for well-conditioned problems.

    See also
    --------
    :class:`toppra.solverwrapper.seidelWrapper`
    :class:`toppra.solverwrapper.hotqpOASESSolverWrapper`

    """

    def _forward_step(self, i, x, K_next):
        """ Compute the highest possible path velocity that is controllable.

        Parameters
        ----------
        i: int
            Current stage index
        x: float
            The squared velocity at the current stage.
        K_next: list
            The lower and upper bounds of the set of controllable squared velocities
            in the next stage.
        eps: float, optional
            A numerical constant to avoid propagating numerical errors.

        Returns
        -------
        optim_var: array
            Optimal variable, which has this format (u, x, v).
            If this step fails, `optim_var` contains only nans.
        """
        # Immediate return
        if None in K_next or i < 0 or i > self._N or np.isnan(x) or x is None:
            return np.array([np.nan, np.nan])

        nV = self.solver_wrapper.get_no_vars()
        g_upper = np.zeros(nV)
        g_upper[1] = -1
        g_upper[0] = -2 * self.solver_wrapper.get_deltas()[i]

        # Account for propagating numerical errors
        K_next_max = K_next[1]
        K_next_min = K_next[0]

        optim_var = self.solver_wrapper.solve_stagewise_optim(
            i, None, g_upper, x, x, K_next_min, K_next_max
        )
        return optim_var



"""
toppra.algorithm.algorithm
^^^^^^^^^^^^^^^^^^^^^^^^^^

This module defines the abstract data types that define TOPP algorithms.

"""
from typing import Dict, Any, List, Tuple, Optional
import typing as T
import abc
import enum
import numpy as np
import time
import matplotlib.pyplot as plt

from toppra.constants import TINY
from toppra.interpolator import SplineInterpolator, AbstractGeometricPath
from toppra.constraint import Constraint
import toppra.interpolator as interpolator
import toppra.parametrizer as tparam

import logging

logger = logging.getLogger(__name__)


class ParameterizationData(object):
    """Internal data and output.
    """
    def __init__(self, *arg, **kwargs) -> None:
        self.return_code: ParameterizationReturnCode = ParameterizationReturnCode.ErrUnknown
        "ParameterizationReturnCode: Return code of the last parametrization attempt."
        self.gridpoints: Optional[np.ndarray] = None
        "np.ndarray: Shape (N+1, 1). Gridpoints"
        self.sd_vec: Optional[np.ndarray] = None
        "np.ndarray: Shape (N+1, 1). Path velocities"
        self.sdd_vec: Optional[np.ndarray] = None
        "np.ndarray: Shape (N+1, 1). Path acceleration"
        self.K: Optional[np.ndarray] = None
        "np.ndarray: Shape (N+1, 2). Controllable sets."
        self.X: Optional[np.ndarray] = None
        "np.ndarray: Shape (N+1, 2). Feasible sets."

    def __repr__(self):
        return "ParameterizationData(return_code:={}, N={:d})".format(
            self.return_code, self.gridpoints.shape[0])


class ParameterizationReturnCode(enum.Enum):
    """Return codes from a parametrization attempt.
    """
    Ok = "Ok: Successful parametrization"
    ErrUnknown = "Error: Unknown issue"
    ErrShortPath = "Error: Input path is very short"
    FailUncontrollable = "Error: Instance is not controllable"
    ErrForwardPassFail = "Error: Forward pass fail. Numerical errors occured"

    def __repr__(self):
        return super(ParameterizationReturnCode, self).__repr__()

    def __str__(self):
        return super(ParameterizationReturnCode, self).__repr__()


class ParameterizationAlgorithm(object):
    """Base parametrization algorithm class.

    This class specifies the generic behavior for parametrization algorithms.  For details on how
    to *construct* a :class:`ParameterizationAlgorithm` instance, as well as configure it, refer
    to the specific class.

    Example usage:

    .. code-block:: python

      # usage
      instance.compute_parametrization(0, 0)
      output = instance.problem_data

      # do this if you only want the final trajectory
      traj = instance.compute_trajectory(0, 0)

    .. seealso::

        :class:`toppra.algorithm.TOPPRA`,
        :class:`toppra.algorithm.TOPPRAsd`,
        :class:`~ParameterizationReturnCode`,
        :class:`~ParameterizationData`

    """

    def __init__(self, constraint_list, path, gridpoints=None, parametrizer=None,
                 gridpt_max_err_threshold: float=1e-3, gridpt_min_nb_points: int=100):
        self.constraints = constraint_list
        self.path = path  # Attr
        self._problem_data = ParameterizationData()
        # Handle gridpoints
        if gridpoints is None:
            gridpoints = interpolator.propose_gridpoints(
                path,
                max_err_threshold=gridpt_max_err_threshold,
                min_nb_points=gridpt_min_nb_points
            )
            logger.info(
                "No gridpoint specified. Automatically choose a gridpoint with %d points",
                len(gridpoints)
            )

        if (
            path.path_interval[0] != gridpoints[0]
            or path.path_interval[1] != gridpoints[-1]
        ):
            raise ValueError("Invalid manually supplied gridpoints.")
        self.gridpoints = np.array(gridpoints)
        self._problem_data.gridpoints = np.array(gridpoints)
        self._N = len(gridpoints) - 1  # Number of stages. Number of point is _N + 1
        for i in range(self._N):
            if gridpoints[i + 1] <= gridpoints[i]:
                logger.fatal("Input gridpoints are not monotonically increasing.")
                raise ValueError("Bad input gridpoints.")
        if parametrizer is None or parametrizer == "ParametrizeSpline":
            # TODO: What is the best way to type parametrizer?
            self.parametrizer: T.Any = tparam.ParametrizeSpline
        elif parametrizer == "ParametrizeConstAccel":
            self.parametrizer = tparam.ParametrizeConstAccel

    @property
    def constraints(self) -> List[Constraint]:
        """Constraints of interests."""
        return self._constraints

    @constraints.setter
    def constraints(self, value: List[Constraint]) -> None:
        # TODO: Validate constraints.
        self._constraints = value

    @property
    def problem_data(self) -> ParameterizationData:
        """Data obtained when solving the path parametrization."""
        return self._problem_data

    @abc.abstractmethod
    def compute_parameterization(self, sd_start: float, sd_end: float, return_data: bool=False):
        """Compute the path parameterization subject to starting and ending conditions.

        After this method terminates, the attribute
        :attr:`~problem_data` will contain algorithm output, as well
        as the result. This is the preferred way of retrieving problem
        output.

        Parameters
        ----------
        sd_start:
            Starting path velocity. Must be positive.
        sd_end:
            Goal path velocity. Must be positive.
        return_data:
            If true also return the problem data.

        """
        raise NotImplementedError

    def compute_trajectory(self, sd_start: float = 0, sd_end: float = 0) -> Optional[AbstractGeometricPath]:
        """Compute the resulting joint trajectory and auxilliary trajectory.

        This is a convenient method if only the final output is wanted.

        Parameters
        ----------
        sd_start:
            Starting path velocity.
        sd_end:
            Goal path velocity.
        return_data:
            If true, return a dict containing the internal data.

        Returns
        -------
        :
            Time-parameterized joint position trajectory or
            None If unable to parameterize. 

        """
        t0 = time.time()
        self.compute_parameterization(sd_start, sd_end)
        if self.problem_data.return_code != ParameterizationReturnCode.Ok:
            logger.warning("Fail to parametrize path. Return code: %s", self.problem_data.return_code)
            return None

        outputtraj = self.parametrizer(self.path, self.problem_data.gridpoints, self.problem_data.sd_vec)
        logger.info("Successfully parametrize path. Duration: %.3f, previously %.3f)",
                    outputtraj.path_interval[1], self.path.path_interval[1])
        logger.info("Finish parametrization in %.3f secs", time.time() - t0)
        return outputtraj

    def inspect(self, compute=True):
        """Inspect the problem internal data."""
        K = self.problem_data.K
        X = self.problem_data.X
        if X is not None:
            plt.plot(X[:, 0], c="green", label="Feasible sets")
            plt.plot(X[:, 1], c="green")
        if K is not None:
            plt.plot(K[:, 0], "--", c="red", label="Controllable sets")
            plt.plot(K[:, 1], "--", c="red")
        if self.problem_data.sd_vec is not None:
            plt.plot(self.problem_data.sd_vec ** 2, label="Velocity profile")
        plt.title("Path-position path-velocity plot")
        plt.xlabel("Path position")
        plt.ylabel("Path velocity square")
        plt.legend()
        plt.tight_layout()
        plt.show()





"""
Some utility functions need to generate PathConstraints. Most are
specific to different scenarios.
"""
import logging
import functools
import warnings

import numpy as np


logger = logging.getLogger(__name__)


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    # pylint: disable=C0111
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn(
            f"Call to deprecated function {func.__name__} in module {func.__module__}.",
            category=DeprecationWarning,
        )
        return func(*args, **kwargs)

    return new_func


def setup_logging(level="WARN"):
    """Setup basic logging facility to console.
    """
    logger_toppra = logging.getLogger("toppra")
    handler_basic = logging.StreamHandler()
    handler_basic.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)5s [%(filename)s : %(lineno)d] %(message)s")
    handler_basic.setFormatter(formatter)
    logger_toppra.setLevel(level)
    logger_toppra.addHandler(handler_basic)


def compute_jacobian_wrench(robot, link, point):
    """Compute the wrench Jacobian for `link` at `point`.

    We look for J_wrench such that
          J_wrench.T * wrench = J_trans.T * F + J_rot.T * tau
    return the induced generalized joint torques.

    J_wrench is computed by stacking J_translation and J_rotation

    """
    jacobian_translation = robot.ComputeJacobianTranslation(link.GetIndex(), point)
    jacobian_rotation = robot.ComputeJacobianAxisAngle(link.GetIndex())
    jacobian_wrench = np.vstack((jacobian_translation, jacobian_rotation))
    return jacobian_wrench


def inv_dyn(rave_robot, q, qd, qdd, forceslist=None, returncomponents=True):
    """Inverse dynamics equation.

    Simple wrapper around OpenRAVE's ComputeInverseDynamics
    function. Return the numerical values of the components of the
    inverse dynamics equation.

          M(q) qdd + C(q, qd) qd + g(q)
        = t1 + t2 + t3

    Parameters
    ----------
    rave_robot : OpenRAVE.robot
    q : (_N, ) ndarray
        Joint position.
    qd : (_N, ) ndarray
        Joint velocity.
    qdd : (_N, ) ndarray
        Joint acceleration.
    returncomponents : Bool
        If True, return the list [t1, t2, t3]
        If False, return t1 + t2 + t3

    Returns
    -------
    res : (3, ) List, or ndarray
        See returncomponents parameter.
    """
    if np.isscalar(q):  # Scalar case
        _q = [q]
        _qd = [qd]
        _qdd = [qdd]
    else:
        _q = q
        _qd = qd
        _qdd = qdd

    # Temporary remove kinematic Limits
    vlim = rave_robot.GetDOFVelocityLimits()
    alim = rave_robot.GetDOFAccelerationLimits()
    rave_robot.SetDOFVelocityLimits(100 * vlim)
    rave_robot.SetDOFAccelerationLimits(100 * alim)
    # Do computation
    with rave_robot:
        rave_robot.SetDOFValues(_q)
        rave_robot.SetDOFVelocities(_qd)
        res = rave_robot.ComputeInverseDynamics(
            _qdd, forceslist, returncomponents=returncomponents
        )
    # Restore kinematic limits
    rave_robot.SetDOFVelocityLimits(vlim)
    rave_robot.SetDOFAccelerationLimits(alim)
    return res


def smooth_singularities(parametrization_instance, us, xs, vs=None):
    """Smooth jitters due to singularities.

    Solving TOPP for discrete problem generated from collocation
    scheme tends to create jitters. This function finds and smooths
    them.

    Notes
    -----
    (`us_smth`, `xs_smth`) is a *valid* path-parameterization. They
    satisfy the linear continuity condition :math:`x_{i+1} = x_i + 2 delta_i u_i`.

    This function is safe: it will always return a solution.

    Parameters
    ----------
    parametrization_instance: :class:`.qpOASESPPSolver`
    us: array
        Shape (_N, ). Controls.
    xs: array
        Shape (_N+1, ). Squared velocities.
    vs: array, optional
        If not given, `vs_smth` will not be returned.

    Returns
    -------
    us_smth: array
        Shape (_N, ). Smoothed controls.
    xs_smth: array
        Shape (_N+1, ). Smoothed squared velocities.
    vs_smth: array
        If `vs` is not given, `vs_smth` will not be returned.

    """
    # Find the indices
    singular_indices = []
    uds = np.diff(us, n=1)
    for i in range(parametrization_instance.N - 3):
        if uds[i] < 0 < uds[i + 1] and uds[i + 2] < 0:
            logger.debug("Found potential singularity at %d", i)
            singular_indices.append(i)
    logger.debug("Found singularities at %s", singular_indices)

    # Smooth the singularities
    xs_smth = np.copy(xs)
    us_smth = np.copy(us)
    if vs is not None:
        vs_smth = np.copy(vs)
    for index in singular_indices:
        idstart = max(0, index)
        idend = min(parametrization_instance.N, index + 4)
        xs_smth[range(idstart, idend + 1)] = xs_smth[idstart] + (
            xs_smth[idend] - xs_smth[idstart]
        ) * np.linspace(0, 1, idend + 1 - idstart)
        if vs is not None:
            data = [
                vs_smth[idstart] + (xs_smth[idend] - xs_smth[idstart]) * frac
                for frac in np.linspace(0, 1, idend + 1 - idstart)
            ]
            vs_smth[range(idstart, idend + 1)] = np.array(data)

    for i in range(parametrization_instance.N):
        us_smth[i] = (
            (xs_smth[i + 1] - xs_smth[i])
            / 2
            / (parametrization_instance.ss[i + 1] - parametrization_instance.ss[i])
        )

    if vs is not None:
        return us_smth, xs_smth, vs_smth
    return us_smth, xs_smth



"""
toppra.solverwrapper
---------------------------

All computations in TOPP-RA algorithms are done by the linear and
quadratic solvers, wrapped in solver wrappers.


SolverWrapper
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.solverwrapper.SolverWrapper
   :members: 


hotqpOASESSolverWrapper
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.solverwrapper.hotqpOASESSolverWrapper
   :members: close_solver, setup_solver, solve_stagewise_optim

qpOASESSolverWrapper
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.solverwrapper.qpOASESSolverWrapper
   :members: close_solver, setup_solver, solve_stagewise_optim


seidelWrapper
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.solverwrapper.seidelWrapper	       
   :members: solve_stagewise_optim


ecosWrapper
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.solverwrapper.ecosWrapper
   :members: solve_stagewise_optim

"""

from .hot_qpoases_solverwrapper import hotqpOASESSolverWrapper
from .cy_seidel_solverwrapper import seidelWrapper
from .ecos_solverwrapper import ecosWrapper
from .qpoases_solverwrapper import qpOASESSolverWrapper
from .solverwrapper import available_solvers, SolverWrapper



"""
Some constants used by TOPPRA solvers.
"""
import logging

try:
    import openravepy as orpy
    FOUND_OPENRAVE = True
except (ImportError, SyntaxError) as err:
    FOUND_OPENRAVE = False
    logging.getLogger("toppra").debug("Unable to import openrave.")


# Constants
SUPERTINY = 1e-10
TINY = 1e-8
SMALL = 1e-5
LARGE = 1000.0
VERYLARGE = 1e8
INFTY = 1e16

# Number of times xs[i] is lowered during the forward pass to
# accomodate for numerical error from solvers.
MAX_TRIES = 10

# TODO: What are these constant used for?
MAXU = 10000  # Max limit for `u`
MAXX = 10000  # Max limit for `x`
MAXSD = 100  # square root of maxx

# constraint creation
JVEL_MAXSD = 1e8  # max sd when creating joint velocity constraints
JACC_MAXU = 1e16  # max u when creating joint acceleration constraint

# solver wrapper related constants.
# NOTE: Read the wrapper's documentation for more details.

# qpoases
QPOASES_INFTY = 1e16

# cvxpy
CVXPY_MAXX = 10000
CVXPY_MAXU = 10000

# ecos
ECOS_MAXX = 10000
ECOS_INFTY = 1000



from setuptools import setup, Extension
from distutils.command.install import install
from Cython.Distutils import build_ext
from Cython.Build import cythonize
import numpy as np
import sys

NAME = "toppra"
with open("VERSION", "r", encoding='UTF-8') as file_:
    VERSION = file_.read()
DESCR = "toppra: time-optimal parametrization of trajectories for robots subject to constraints."
with open("README.md", "r", encoding='UTF-8') as file_:
    LONG_DESCRIPTION = file_.read()

URL = "https://github.com/hungpham2511/toppra"


# setup requirements
if sys.version[0] == '2':
    with open("requirements.txt", "r") as f:
        REQUIRES = ["scipy==0.18.0", "numpy", "matplotlib",
                    # only required on python2.7
                    "pathlib2", "enum34", "strip_hints", "typing"]
        DEV_REQUIRES = [line.strip() for line in f if line.strip()]
else:
    with open("requirements3.txt", "r") as f:
        REQUIRES = ["scipy>0.18", "numpy", "matplotlib"]
        DEV_REQUIRES = [line.strip() for line in f if line.strip()]

AUTHOR = "Hung Pham"
EMAIL = "hungpham2511@gmail.com"

LICENSE = "MIT"

SRC_DIR = "toppra"
PACKAGES = ["toppra",
            "toppra.constraint",
            "toppra.algorithm",
            "toppra.algorithm.reachabilitybased",
            "toppra.solverwrapper",
            "toppra.cpp"]

ext_1 = Extension(SRC_DIR + "._CythonUtils",
                  [SRC_DIR + "/_CythonUtils.pyx"],
                  extra_compile_args=['-O1'],
                  libraries=[],
                  include_dirs=[np.get_include()])

ext_2 = Extension(SRC_DIR + ".solverwrapper.cy_seidel_solverwrapper",
                  [SRC_DIR + "/solverwrapper/cy_seidel_solverwrapper.pyx"],
                  extra_compile_args=['-O1'],
                  include_dirs=[np.get_include()])

EXTENSIONS = [ext_1, ext_2]
SETUP_REQUIRES = ["numpy", "cython"]
if sys.version[0] == '2' or sys.version[:3] == '3.5':
    SETUP_REQUIRES = ["numpy", "cython", "strip_hints"]


# custom install command: strip type-hints before installing toppra
# for python2.7 and pthon3.5
class install2(install):
    def run(self, *args, **kwargs):
        # stripping
        if sys.version[0] == '2' or sys.version[:3] == '3.5':
            from strip_hints import strip_file_to_string
            import glob
            import os.path
            def process_file(f):
                print(os.path.abspath(f))
                out = strip_file_to_string(f)
                with open(f, 'w') as fh:
                    fh.write(out)
            for f in glob.glob("%s/*/toppra/*/*.py" % self.build_base):
                process_file(f)
            for f in glob.glob("%s/*/toppra/*.py" % self.build_base):
                process_file(f)

            print(os.path.abspath("."))
            print(os.path.abspath(self.build_base))
        # install new files
        install.run(self, *args, **kwargs)


if __name__ == "__main__":
    setup(
        # Dependencies installed when running `pip install .`
        install_requires=REQUIRES,
        setup_requires=["numpy", "cython"],
        extras_require={
            # Dependencies installed when running `pip install -e .[dev]`
            
            # NOTE: This is deprecated in favour of the simpler workflow
            # of installing from requirements3.txt before installing
            # this pkg.
            'dev': DEV_REQUIRES  
        },
        packages=PACKAGES,
        zip_safe=False,
        name=NAME,
        version=VERSION,
        description=DESCR,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        author=AUTHOR,
        author_email=EMAIL,
        url=URL,
        license=LICENSE,

          # This is used to build the Cython modules. Will be run
        # automatically if not found by pip. Otherwise run
        #
        #      python setup.py build
        #
        # to trigger manually.
          cmdclass={"build_ext": build_ext, "install": install2},
        ext_modules=cythonize(EXTENSIONS)
    )



from .linear_constraint import LinearConstraint, canlinear_colloc_to_interpolate
from ..constraint import DiscretizationType
import numpy as np


class JointTorqueConstraint(LinearConstraint):
    """Joint Torque Constraint.

    A joint torque constraint is given by

    .. math::
        A(q) \ddot q + \dot q^\\top B(q) \dot q + C(q) + D( \dot q )= w,

    where w is a vector that satisfies the polyhedral constraint:

    .. math::
        F(q) w \\leq g(q).

    Notice that `inv_dyn(q, qd, qdd) = w` and that `cnsf_coeffs(q) =
    F(q), g(q)`.

    To evaluate the constraint on a geometric path `p(s)`, multiple
    calls to `inv_dyn` and `const_coeff` are made. Specifically one
    can derive the second-order equation as follows

    .. math::
        A(q) p'(s) \ddot s + [A(q) p''(s) + p'(s)^\\top B(q) p'(s)] \dot s^2 + C(q) + D( \dot q ) = w,
        a(s) \ddot s + b(s) \dot s ^2 + c(s) = w

    To evaluate the coefficients a(s), b(s), c(s), inv_dyn is called
    repeatedly with appropriate arguments.

    Parameters
    ----------

    inv_dyn: (array, array, array) -> array
        The "inverse dynamics" function that receives joint position, velocity and
        acceleration as inputs and ouputs the "joint torque". See notes for more
        details.

    tau_lim: array
        Shape (dof, 2). The lower and upper torque bounds of the
        j-th joint are tau_lim[j, 0] and tau_lim[j, 1] respectively.

    fs_coef: array
        Shape (dof). The coefficients of dry friction of the
        joints.

    discretization_scheme: :class:`.DiscretizationType`
        Can be either Collocation (0) or Interpolation
        (1). Interpolation gives more accurate results with slightly
        higher computational cost.

    """

    def __init__(
        self,
        inv_dyn,
        tau_lim,
        fs_coef,
        discretization_scheme=DiscretizationType.Collocation,
    ):
        super(JointTorqueConstraint, self).__init__()
        self.inv_dyn = inv_dyn
        self.tau_lim = np.array(tau_lim, dtype=float)
        self.fs_coef = np.array(fs_coef)
        self.dof = self.tau_lim.shape[0]
        self.set_discretization_type(discretization_scheme)
        assert self.tau_lim.shape[1] == 2, "Wrong input shape."
        self._format_string = "    Torque limit: \n"
        for i in range(self.tau_lim.shape[0]):
            self._format_string += (
                "      J{:d}: {:}".format(i + 1, self.tau_lim[i]) + "\n"
            )
        self.identical = True

    def compute_constraint_params(self, path, gridpoints):
        if path.dof != self.get_dof():
            raise ValueError(
                "Wrong dimension: constraint dof ({:d}) not equal to path dof ({:d})".format(
                    self.get_dof(), path.dof
                )
            )
        v_zero = np.zeros(path.dof)
        p = path.eval(gridpoints)
        ps = path.evald(gridpoints)
        pss = path.evaldd(gridpoints)
        N = gridpoints.shape[0] - 1
        dof = path.dof
        I_dof = np.eye(dof)
        F = np.zeros((dof * 2, dof))
        g = np.zeros(dof * 2)
        g[0:dof] = self.tau_lim[:, 1]
        g[dof:] = -self.tau_lim[:, 0]
        F[0:dof, :] = I_dof
        F[dof:, :] = -I_dof

        c = np.array([self.inv_dyn(p_, v_zero, v_zero) for p_ in p])
        a = np.array([self.inv_dyn(p_, v_zero, ps_) for p_, ps_ in zip(p, ps)]) - c
        b = (
            np.array([self.inv_dyn(p_, ps_, pss_) for p_, ps_, pss_ in zip(p, ps, pss)])
            - c
        )

        # dry friction
        for i in range(0, dof):
            c[:, i] += self.fs_coef[i] * np.sign(ps[:, i])

        if self.discretization_type == DiscretizationType.Collocation:
            return a, b, c, F, g, None, None
        elif self.discretization_type == DiscretizationType.Interpolation:
            return canlinear_colloc_to_interpolate(
                a, b, c, F, g, None, None, gridpoints, identical=True
            )
        else:
            raise NotImplementedError("Other form of discretization not supported!")



"""
Constraints
----------------------

Modules implementing different dynamics constraints.

ConstraintType
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.constraint.ConstraintType
   :members:

DiscretizationType
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.constraint.DiscretizationType
   :members:

LinearConstraint
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.constraint.LinearConstraint
   :members:
   :special-members:
   :show-inheritance:


JointVelocityConstraint
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.constraint.JointVelocityConstraint
   :members:
   :special-members:
   :show-inheritance:

JointVelocityConstraintVarying
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.constraint.JointVelocityConstraintVarying
   :members:
   :special-members:
   :show-inheritance:

SecondOrderConstraints
^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.constraint.SecondOrderConstraint
   :members:
   :special-members:
   :show-inheritance:

JointTorqueConstraint
^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.constraint.JointTorqueConstraint
   :members:
   :special-members:
   :show-inheritance:

JointAccelerationConstraint
^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.constraint.JointAccelerationConstraint
   :members:
   :special-members:
   :show-inheritance:

RobustLinearConstraint
^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autoclass:: toppra.constraint.RobustLinearConstraint
   :members:
   :show-inheritance:

[abstract]Constraint
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: toppra.constraint.Constraint
   :members: compute_constraint_params, set_discretization_type, __repr__

[internal]Methods
^^^^^^^^^^^^^^^^^^^^^^^^
.. autofunction:: toppra.constraint.canlinear_colloc_to_interpolate

"""
from .constraint import ConstraintType, DiscretizationType, Constraint
from .joint_torque import JointTorqueConstraint
from .linear_joint_acceleration import JointAccelerationConstraint
from .linear_joint_velocity import (
    JointVelocityConstraint,
    JointVelocityConstraintVarying,
)
from .linear_second_order import SecondOrderConstraint, canlinear_colloc_to_interpolate
from .conic_constraint import RobustLinearConstraint
from .linear_constraint import LinearConstraint



"""This module implements the joint velocity constraint."""
import numpy as np
from toppra._CythonUtils import (_create_velocity_constraint,
                                 _create_velocity_constraint_varying)
from .linear_constraint import LinearConstraint


class JointVelocityConstraint(LinearConstraint):
    """A Joint Velocity Constraint class.

    Parameters
    ----------
    vlim: np.ndarray
        Shape (dof, 2). The lower and upper velocity bounds of the j-th joint
        are given by vlim[j, 0] and vlim[j, 1] respectively.

    """

    def __init__(self, vlim):
        super(JointVelocityConstraint, self).__init__()
        vlim = np.array(vlim, dtype=float)
        if np.isnan(vlim).any():
            raise ValueError("Bad velocity given: %s" % vlim)
        if len(vlim.shape) == 1:
            self.vlim = np.vstack((-np.array(vlim), np.array(vlim))).T
        else:
            self.vlim = np.array(vlim, dtype=float)
        self.dof = self.vlim.shape[0]
        self._assert_valid_limits()

    def _assert_valid_limits(self):
        """Check that the velocity limits is valid."""
        assert self.vlim.shape[1] == 2, "Wrong input shape."
        for i in range(self.dof):
            if self.vlim[i, 0] >= self.vlim[i, 1]:
                raise ValueError("Bad velocity limits: {:} (lower limit) > {:} (higher limit)".format(
                    self.vlim[i, 0], self.vlim[i, 1]))
        self._format_string = "    Velocity limit: \n"
        for i in range(self.vlim.shape[0]):
            self._format_string += "      J{:d}: {:}".format(
                i + 1, self.vlim[i]) + "\n"

    def compute_constraint_params(self, path, gridpoints):
        if path.dof != self.get_dof():
            raise ValueError(
                "Wrong dimension: constraint dof ({:d}) not equal to path dof ({:d})"
                .format(self.get_dof(), path.dof))
        qs = path(gridpoints, 1)
        _, _, xbound_ = _create_velocity_constraint(qs, self.vlim)
        xbound = np.array(xbound_)
        xbound[:, 0] = xbound_[:, 1]
        xbound[:, 1] = -xbound_[:, 0]
        return None, None, None, None, None, None, xbound


class JointVelocityConstraintVarying(LinearConstraint):
    """A Joint Velocity Constraint class.

    This class handle velocity constraints that vary along the path.

    Parameters
    ----------
    vlim_func: (float) -> np.ndarray
        A function that receives a scalar (float) and produce an array
        with shape (dof, 2). The lower and upper velocity bounds of
        the j-th joint are given by out[j, 0] and out[j, 1]
        respectively.
    """

    def __init__(self, vlim_func):
        super(JointVelocityConstraintVarying, self).__init__()
        self.dof = vlim_func(0).shape[0]
        self._format_string = "    Varying Velocity limit: \n"
        self.vlim_func = vlim_func

    def compute_constraint_params(self, path, gridpoints):
        if path.dof != self.get_dof():
            raise ValueError(
                "Wrong dimension: constraint dof ({:d}) not equal to path dof ({:d})"
                .format(self.get_dof(), path.dof))
        qs = path((gridpoints), 1)
        vlim_grid = np.array([self.vlim_func(s) for s in gridpoints])
        _, _, xbound_ = _create_velocity_constraint_varying(qs, vlim_grid)
        xbound = np.array(xbound_)
        xbound[:, 0] = xbound_[:, 1]
        xbound[:, 1] = -xbound_[:, 0]
        return None, None, None, None, None, None, xbound



"""This module provides different solverwrapper implementations."""
import logging
import numpy as np

logger = logging.getLogger(__name__)


# pylint: disable=unused-import
def available_solvers(output_msg=True):
    """Check for available solvers."""
    try:
        import ecos

        IMPORT_ECOS = True
    except ImportError as err:
        IMPORT_ECOS = False
    try:
        import qpoases

        IMPORT_QPOASES = True
    except ImportError as err:
        IMPORT_QPOASES = False
    try:
        import cvxpy

        IMPORT_CVXPY = True
    except ImportError as err:
        IMPORT_CVXPY = False
    solver_availability = (
        ("seidel", True),
        ("hotqpoases", IMPORT_QPOASES),
        ("qpoases", IMPORT_QPOASES),
        ("ecos", IMPORT_ECOS),
        ("cvxpy", IMPORT_CVXPY),
    )

    if output_msg:
        print(solver_availability)
    return solver_availability

def check_solver_availability(solver):
    check = available_solvers(False)
    for sname, avail in check:
        if sname == solver and avail:
            return True
    return False


class SolverWrapper(object):
    """The base class for all solver wrappers.

    All SolverWrapper have to implement a core method needed by all
    Reachability Analysis-based algorithms:
    `solve_stagewise_optim`. This methods solves a Linear/Quadratic
    Program subject to linear constraints at the given stage, and
    possibly with additional auxiliary constraints.

    Note that some SolverWrappers only handle Linear Program while
    some handle both.

    Certain solver wrappers need to be setup and close down before and
    after usage. For instance, the wrappers for mosek and qpOASES with
    warmstart capability. To provide this functionality, this class
    contains two abstract methods `setup_solver` and `close_solver`,
    which should be called before and after any call to
    `solve_stagewise_optim`, so that necessary setups can be made.

    Each solver wrapper should provide solver-specific constraint,
    such as ultimate bound the variable u, x. For some solvers such as
    ECOS, this is very important.

    Attributes
    ----------
    constraints : :class:`toppra.Constraint` []
        Constraints on the robot system.
    path : Interpolator
        The geometric path to be time-parametrized.
    path_discretization: array
        The discretization grid use to discretize the geometric path.

    """

    def __init__(self, constraint_list, path, path_discretization):
        # Main attributes
        self.constraints = constraint_list
        self.path = path
        self.path_discretization = np.array(path_discretization)
        # End main attributes
        self.N = (
            len(path_discretization) - 1
        )  # Number of stages. Number of point is _N + 1
        self.deltas = self.path_discretization[1:] - self.path_discretization[:-1]
        for i in range(self.N):
            assert path_discretization[i + 1] > path_discretization[i]

        self.params = [
            c.compute_constraint_params(
                self.path, self.path_discretization
            )
            for c in self.constraints
        ]
        self.nV = 2 + sum([c.get_no_extra_vars() for c in self.constraints])

    def get_no_stages(self):
        """Return the number of stages.

        The number of gridpoints equals N + 1, where N is the number
        of stages.
        """
        return self.N

    def get_no_vars(self):
        """ Return total number of variables, including u, x.
        """
        return self.nV

    def get_deltas(self):
        return self.deltas

    def solve_stagewise_optim(self, i, H, g, x_min, x_max, x_next_min, x_next_max):
        """Solve a stage-wise quadratic (or linear) optimization problem.

        The quadratic optimization problem is described below:

        .. math::
            \\text{min  }  & 0.5 [u, x, v] H [u, x, v]^\\top + [u, x, v] g    \\\\
            \\text{s.t.  } & [u, x] \\text{ is feasible at stage } i \\\\
                           & x_{min} \leq x \leq x_{max}             \\\\
                           & x_{next, min} \leq x + 2 \Delta_i u \leq x_{next, max},

        where `v` is an auxiliary variable, only exist if there are
        non-canonical constraints.  The linear program is the
        quadratic problem without the quadratic term.

        Parameters
        ----------
        i: int
            The stage index.
        H: :class:`~numpy:numpy.ndarray` or None
            Shape (d, d). The coefficient of the quadratic objective
            function. If is None, set the quadratic term to zero.
        g: :class:`~numpy:numpy.ndarray`
            Shape (d,). The linear term.
        x_min: float
            If not specified, set to NaN.
        x_max: float
            If not specified, set to NaN.
        x_next_min: float
            If not specified, set to NaN.
        x_next_max: float
            If not specified, set to NaN.

        Returns
        -------
        double :class:`~numpy:numpy.ndarray`
             If successes, return a double array containing the optimal
             variable.  Otherwise return a array that contains `numpy.nan`.

        """
        raise NotImplementedError

    def setup_solver(self):
        pass

    def close_solver(self):
        pass
